@@ -25,15 +25,13 @@ knight_warps = [(2, 1), (2, -1), (-2, 1), (-2, -1), (1, 2), (-1, 2), (1, -2), (-
 king_warps = [(1, 1), (-1, -1), (1, -1), (-1, 1), (0, 1), (0, -1), (1, 0), (-1, 0)]
 pawn_capture_warps = [[(-1, 1), (1, 1)], [(-1, -1), (1, -1)]]
 EMPTY, WHITE, BLACK = 0, 1, 2
-NORMAL, CHECK, CHECKMATE, STALEMATE = 0, 1, 2, 3
+NORMAL, CHECK, CHECKMATE, STALEMATE, DRAW = 0, 1, 2, 3, 4
 
 def file_to_index(file): return ord(file) - 97
 def index_to_file(index): return chr(index + 97)
 def rank_file_to_coord(to_rank_file): return (file_to_index(to_rank_file[0]), int(to_rank_file[1]) - 1)
 def coord_to_notation(coord): return index_to_file(coord[0]) + str(coord[1]+1)
-def in_range(coord): return coord[0] in range(SIZE) and coord[1] in range(SIZE)
 def inverse_color(color): return WHITE if color == BLACK else BLACK
-def map2(func, vals): return [func(val) for val in vals] #map with side effects
 def threatened_str(threatened):
     piece_arr = [str(threatened[coord[1]][SIZE - coord[0] - 1]) for coord in all_coords]
     return "".join([" ".join(piece_arr[i*SIZE:i*SIZE+SIZE])+"\n" for i in range(SIZE)])
@@ -74,11 +72,11 @@ class Chess:
     def __set_row(self, row, row_pieces):
         for col in range(SIZE):
             self.set_coord((col, row), pieces_index[row_pieces[col]])
-        #map2(lambda col: self.coord((col, row), pieces_index[row_pieces[col]]), [col for col in range(SIZE)])
 
 
     def init_player_pieces(self):
-        self.player_pieces_list = [{c for c in all_coords if self.is_color(c, color)} for color in (WHITE, BLACK)]
+        self.player_pieces_list = [{c for c in all_coords if self.color(c) == color} for color in (WHITE, BLACK)]
+        self.player_pieces_types = {}
 
 
     def set_coord(self, coord: tuple, piece):
@@ -89,11 +87,6 @@ class Chess:
         return self.board[coord[1]][coord[0]]
 
 
-    def is_color(self, coord: tuple, color):
-        """Returns True if the color (player) of the piece at a location is the given color."""
-        return self.color(coord) == color
-
-
     def color(self, coord: tuple):
         """Returns the color of the piece at a location (or the EMPTY color if no piece is at that location)."""
         piece_type = self.board[coord[1]][coord[0]]
@@ -104,13 +97,13 @@ class Chess:
 
     def is_type(self, coord: tuple, piece_type: str):
         """Returns true if the piece at a location is the given type of piece."""
-        return pieces[self.get_coord(coord)].lower() == piece_type.lower()
+        return pieces[self.board[coord[1]][coord[0]]].lower() == piece_type.lower()
 
 
     def player_pieces_of_type(self, piece_type, player):
         """Get all pieces of a certain type for a player."""
         for coord in self.player_pieces_list[player - 1]:
-            if self.get_coord(coord) in pieces_type_map[piece_type]:
+            if self.board[coord[1]][coord[0]] in pieces_type_map[piece_type]:
                 yield coord
 
 
@@ -119,16 +112,48 @@ class Chess:
         return self.player_pieces_list[player - 1]
 
 
+    def get_player_piece_types(self):
+        move_num = len(self.move_list)
+        if move_num not in self.player_pieces_types:
+            white_pieces = [self.board[p[1]][p[0]] for p in self.player_pieces(WHITE)]
+            white_pieces.sort()
+            black_pieces = [self.board[p[1]][p[0]] for p in self.player_pieces(BLACK)]
+            black_pieces.sort()
+            self.player_pieces_types[move_num] = [white_pieces, black_pieces]
+        return self.player_pieces_types[move_num][0], self.player_pieces_types[move_num][1]
+
+
     def game_state(self):
-        #FIXME cache the result of this and invalidate the cache when a move is made
+        #FIXME cache threats and pass it to check_check
         #TODO threefold repetition and fifty move rule?
-        #FIXME check for impossible checkmate (king/king, king/bishop, king/knight, king/bishop v king/bishop of same color
+        num_pieces = len(self.player_pieces_list[0]) + len(self.player_pieces_list[1])
+        if num_pieces <= 4:
+            # check for sufficient material
+            white_pieces, black_pieces = self.get_player_piece_types()
+            if white_pieces == [1] and black_pieces == [7]:  # king v king
+                return DRAW
+            if white_pieces == [1, 4] and black_pieces == [7, 10]: # king bishop v king bishop
+                return DRAW
+            if white_pieces == [1] and black_pieces == [7, 10]:  # king v king bishop
+                return DRAW
+            if white_pieces == [1, 4] and black_pieces == [7]:  # king bishop v king
+                return DRAW
+            if white_pieces == [1, 5] and black_pieces == [7, 11]:  # king knight v king knight
+                return DRAW
+            if white_pieces == [1] and black_pieces == [7, 11]:  # king v king knight
+                return DRAW
+            if white_pieces == [1, 5] and black_pieces == [7]: # king knight v king
+                return DRAW
+            if white_pieces == [1, 5] and black_pieces == [7, 10]:  # king knight v king bishop
+                return DRAW
+            if white_pieces == [1, 4] and black_pieces == [7, 11]:  # king bishop v king knight
+                return DRAW
         piece_moves = self.valid_moves_for_player(self.current_player, True)
-        moves = []
+        has_next_move = False
         for piece in piece_moves:
-            moves.extend(piece_moves[piece])
-        check = self.check_check(self.current_player)
-        if len(moves) == 0:
+            has_next_move = has_next_move or (next(piece_moves[piece], None) is not None)
+        check = self.in_check(self.current_player)
+        if not has_next_move:
             return CHECKMATE if check else STALEMATE
         return CHECK if check else NORMAL
 
@@ -140,22 +165,21 @@ class Chess:
 
 
     def valid_moves_for_player(self, player, validate=True, threats=None):
-        pieces = copy.deepcopy(self.player_pieces(player))
-        return {piece: list(self.valid_piece_moves(piece, validate, threats)) for piece in pieces}
+        pieces = [p for p in self.player_pieces(player)]
+        return {piece: self.valid_piece_moves(piece, validate, threats) for piece in pieces}
 
 
     def valid_piece_moves(self, p, validate=True, threats=None):
-        # TODO detect check and checkmate
-        p_type = self.get_coord(p)
+        p_type = self.board[p[1]][p[0]]
         p_type = p_type - 6 if p_type > 6 else p_type
-        funcs = [lambda: [], self.king, self.queen, self.rook, self.bishop, self.knight, self.pawn]
+        funcs = [lambda f, r: [], self.__king, self.__queen, self.__rook, self.__bishop, self.__knight, self.__pawn]
         moves = funcs[p_type](p[0], p[1], threats) if p_type == 1 else funcs[p_type](p[0], p[1])
         if validate:
             player = self.current_player
             for move in moves: #simulate to prevent moving into check
                 try:
                     self.move(p, move, False, False)
-                    if not self.check_check(player):
+                    if not self.in_check(player, threats=threats):
                         self.__undo_last_move()
                         yield move
                     else:
@@ -205,51 +229,51 @@ class Chess:
                 if move[2]:
                     yield move
         for pawn in self.player_pieces_of_type("P", inverse_color(player)):
-            for threat in self.pawn_threats(pawn[0], pawn[1]):
+            for threat in self.__pawn_threats(pawn[0], pawn[1]):
                 yield threat
 
 
-    def check_check(self, player, threats=None):
+    def in_check(self, player, threats=None):
         """
         Determines if a player is currently in check.
         A precomputed threat matrix can be provided as an optimization.
         """
-        #XXX can we use the optimized threats function to just check the king square?
+        k = list(self.player_pieces_of_type("K", player))[0]
         threats = self.__compute_threats(player) if threats is None else threats
-        king = list(self.player_pieces_of_type("K", player))
-        return (king[0][0], king[0][1], True) in threats
+        return (k[0], k[1], True) in threats
 
 
-    def move(self, from_coord, to_coord, validate=True, pgn_gen=True):
+    def move(self, fromc, to, validate=True, pgn_gen=True):
+        threats = None
         if validate:
-            threats = self.compute_threat_matrix(self.current_player) if self.is_type(from_coord, "K") else None
-            valid_moves = self.valid_piece_moves(from_coord, True, threats) if self.is_color(from_coord, self.current_player) else {}
+            threats = self.compute_threat_matrix(self.current_player) if self.is_type(fromc, "K") else None
+            valid_moves = self.valid_piece_moves(fromc, True, threats) if self.color(fromc) == self.current_player else {}
             #below line used for debugging valid moves only
             #valid_moves = list(valid_moves)
-        if not validate or to_coord in valid_moves:
+        if not validate or to in valid_moves:
             if pgn_gen and validate:
-                self.__append_move_to_pgn(from_coord, pgn_gen, to_coord, validate)
-            if self.is_type(from_coord, "R") and from_coord in rook_positions_index:
-                rook_index = rook_positions_index[from_coord]
+                self.__append_move_to_pgn(fromc, pgn_gen, to, validate, threats)
+            if self.is_type(fromc, "R") and fromc in rook_positions_index:
+                rook_index = rook_positions_index[fromc]
                 if self.rooks_moved[rook_index] == -1:
                     self.rooks_moved[rook_index] = len(self.move_list)
-            elif self.is_type(from_coord, "K") and from_coord in king_positions:
-                king_index = king_positions.index(from_coord)
+            elif self.is_type(fromc, "K") and fromc in king_positions:
+                king_index = king_positions.index(fromc)
                 on_move = len(self.move_list)
                 if self.kings_moved[king_index] == -1:
                     self.kings_moved[king_index] = on_move
-                if to_coord in king_castle_end_positions and self.kings_moved[king_index] == on_move:
-                    rook_index = king_castle_end_positions_index[to_coord]
+                if to in king_castle_end_positions and self.kings_moved[king_index] == on_move:
+                    rook_index = king_castle_end_positions_index[to]
                     rook_position = rook_positions[rook_index]
-                    if self.is_type(rook_position, "R") and self.is_color(rook_position, self.current_player):
+                    if self.is_type(rook_position, "R") and self.color(rook_position) == self.current_player:
                         self.rooks_moved[rook_index] = on_move
                         self.__move_piece(rook_position, rook_castle_end_positions[rook_index])
-                        castle_pos_index = king_castle_end_positions.index(to_coord)
+                        castle_pos_index = king_castle_end_positions.index(to)
                         self.kings_castled[king_index] = on_move
-            self.__move_piece(from_coord, to_coord)
-            self.move_list.append((from_coord, (to_coord[0], to_coord[1])))
-            if len(to_coord) == 4:
-                self.__handle_special(from_coord, to_coord)
+            self.__move_piece(fromc, to)
+            self.move_list.append((fromc, (to[0], to[1])))
+            if len(to) == 4:
+                self.__handle_special(fromc, to)
             self.current_player = inverse_color(self.current_player)
             if pgn_gen and validate:
                 self.__append_game_state_to_pgn()
@@ -257,34 +281,34 @@ class Chess:
             raise BadMoveException("Move is invalid")
 
 
-    def __append_move_to_pgn(self, from_coord, pgn_gen, to_coord, validate):
+    def __append_move_to_pgn(self, fromc, pgn_gen, to, validate, threats):
         pgn_castle, pgn_promotion, file_disambiguation, rank_disambiguation = None, "", "", ""
-        if self.is_type(from_coord, "K") and from_coord in king_positions:
-            if to_coord in king_castle_end_positions:
-                castle_pos_index = king_castle_end_positions.index(to_coord)
+        if self.is_type(fromc, "K") and fromc in king_positions:
+            if to in king_castle_end_positions:
+                castle_pos_index = king_castle_end_positions.index(to)
                 pgn_castle = "O-O-O " if castle_pos_index == 0 or castle_pos_index == 2 else "O-O "
-        to_color = self.color(to_coord)
+        to_color = self.color(to)
         capture_str = "x" if to_color != EMPTY else ""
         if pgn_castle is not None:
             pgn = pgn_castle
         else:
-            pgn_type = pieces[self.get_coord(from_coord)].upper()
-            piece_type = self.get_coord(from_coord)
+            pgn_type = pieces[self.get_coord(fromc)].upper()
+            piece_type = self.get_coord(fromc)
             # TODO piece disambiguation if same type of piece can move to to_coord
             for piece in self.__type_in_coords(self.player_pieces_list[self.current_player - 1], piece_type):
-                if self.get_coord(piece) == piece_type and piece != from_coord:
-                    for move in self.valid_piece_moves(piece):
-                        if move == to_coord:
-                            if piece[0] == from_coord[0]:
-                                rank_disambiguation = str(from_coord[1] + 1)
+                if self.board[piece[1]][piece[0]] == piece_type and piece != fromc:
+                    for move in self.valid_piece_moves(piece, threats=threats):
+                        if move == to:
+                            if piece[0] == fromc[0]:
+                                rank_disambiguation = str(fromc[1] + 1)
                             else:
-                                file_disambiguation = index_to_file(from_coord[0])
-            notation_coord = coord_to_notation(to_coord)
+                                file_disambiguation = index_to_file(fromc[0])
+            notation_coord = coord_to_notation(to)
             if pgn_type == "P":
-                if len(to_coord) == 4 and to_coord[3] != 6 and to_coord[3] != 12:
-                    pgn_promotion = "=" + to_coord[3].upper()
+                if len(to) == 4 and to[3] != 6 and to[3] != 12:
+                    pgn_promotion = "=" + to[3].upper()
                 if capture_str != "":
-                    pgn = index_to_file(from_coord[0]) + capture_str + notation_coord + pgn_promotion
+                    pgn = index_to_file(fromc[0]) + capture_str + notation_coord + pgn_promotion
                 else:
                     pgn = notation_coord + pgn_promotion
             else:
@@ -314,17 +338,17 @@ class Chess:
                 self.pgn_str.append(game_termination_markers[0])
 
 
-    def __move_piece(self, from_coord, to_coord):
-        from_color = self.color(from_coord)
-        to_color = self.color(to_coord)
+    def __move_piece(self, fromc, to):
+        from_color = self.color(fromc)
+        to_color = self.color(to)
         if to_color != EMPTY:
-            captured = (to_coord[0], to_coord[1])
-            self.captured_pieces[len(self.move_list)] = (self.get_coord(captured), captured)
+            captured = (to[0], to[1])
+            self.captured_pieces[len(self.move_list)] = (self.board[to[1]][to[0]], captured)
             self.player_pieces_list[to_color - 1].remove(captured)
-        self.player_pieces_list[from_color - 1].remove(from_coord)
-        self.player_pieces_list[from_color - 1].add((to_coord[0], to_coord[1]))
-        self.set_coord(to_coord, self.get_coord(from_coord))
-        self.set_coord(from_coord, EMPTY)
+        self.player_pieces_list[from_color - 1].remove(fromc)
+        self.player_pieces_list[from_color - 1].add((to[0], to[1]))
+        self.board[to[1]][to[0]] = self.board[fromc[1]][fromc[0]]
+        self.board[fromc[1]][fromc[0]] = EMPTY
 
 
     def __undo_last_move(self):
@@ -349,46 +373,46 @@ class Chess:
                     self.rooks_moved[rook_index] = -1
                     self.kings_castled[king_index] = -1
         elif last_move_index in self.promoted_pieces:
-            self.set_coord(last_move[1], 6 if self.current_player == WHITE else 12)
+            self.board[last_move[1][1]][last_move[1][0]] = 6 if self.current_player == WHITE else 12
             self.promoted_pieces.pop(last_move_index)
         self.__move_piece(last_move[1], last_move[0])
         if captured is not None:
-            self.set_coord(captured[1], captured[0])
+            self.board[captured[1][1]][captured[1][0]] = captured[0]
             self.player_pieces_list[self.color(captured[1]) - 1].add(captured[1])
 
-    def __handle_special(self, from_coord, to_coord):
+    def __handle_special(self, fromc, to):
         """Handles one-off move types like en pessant and pawn promotion when applying a move."""
-        if to_coord[3] == 6 or to_coord[3] == 12: #en pessant
-            captured = (to_coord[0], from_coord[1])
+        if to[3] == 6 or to[3] == 12: #en pessant
+            captured = (to[0], fromc[1])
             captured_color = self.color(captured)
-            self.captured_pieces[len(self.move_list) - 1] = (self.get_coord(captured), captured)
-            self.set_coord(captured, EMPTY)
+            self.captured_pieces[len(self.move_list) - 1] = (self.board[captured[1]][captured[0]], captured)
+            self.board[captured[1]][captured[0]] = EMPTY
             self.player_pieces_list[captured_color - 1].remove(captured)
         else: #promotion
-            promotion_type = to_coord[3]
-            if promotion_type.upper() in promotion_candidates and self.is_type(to_coord, "P"):
-                if self.current_player == WHITE and to_coord[1] == 7:
-                    self.set_coord(to_coord, pieces_index[promotion_type.upper()])
-                elif self.current_player == BLACK and to_coord[1] == 0:
-                    self.set_coord(to_coord, pieces_index[promotion_type.lower()])
+            promotion_type = to[3]
+            if promotion_type.upper() in promotion_candidates and self.is_type(to, "P"):
+                if self.current_player == WHITE and to[1] == 7:
+                    self.board[to[1]][to[0]] = pieces_index[promotion_type.upper()]
+                elif self.current_player == BLACK and to[1] == 0:
+                    self.board[to[1]][to[0]] = pieces_index[promotion_type.lower()]
                 else:
                     raise BadPromotionException("Cannot promote from this position")
-                self.promoted_pieces[len(self.move_list)-1] = (self.get_coord(to_coord), to_coord)
+                self.promoted_pieces[len(self.move_list)-1] = (self.board[to[1]][to[0]], to)
             else:
                 raise BadPromotionException("Invalid promotion")
 
 
-    def pawn_threats(self, f, r):
+    def __pawn_threats(self, f, r):
         color = self.color((f, r))
         for threat in self.__warps([], (f, r), pawn_capture_warps[color - 1], (color,)):
             yield threat
 
 
-    def pawn(self, f, r):
+    def __pawn(self, f, r):
         """Returns all valid moves for a pawn at file f and rank r including en pessant and valid promotions."""
         color = self.color((f, r))
-        warps = [[(0, 1), (0, 2)] if r == 1 and self.is_color((f, r + 1), EMPTY) else [(0, 1)]]
-        warps.append([(0, -1), (0, -2)] if r == 6 and self.is_color((f, r - 1), EMPTY) else [(0, -1)])
+        warps = [[(0, 1), (0, 2)] if r == 1 and self.color((f, r + 1)) == EMPTY else [(0, 1)]]
+        warps.append([(0, -1), (0, -2)] if r == 6 and self.color((f, r - 1)) == EMPTY else [(0, -1)])
         moves = self.__warps([], (f, r), warps[color - 1], (WHITE, BLACK), False)
         for move in self.__warps(moves, (f, r), pawn_capture_warps[color - 1], (EMPTY, color)):
             if color == WHITE and move[1] == 7:
@@ -401,49 +425,49 @@ class Chess:
                 yield move
         if len(self.move_list) > 0: #en pessant - check that last move was a double pawn advance
             last_move = self.move_list[-1]
-            if self.is_type(last_move[1], "P") and self.is_color(last_move[1], inverse_color(color)):
+            piece_type = self.board[last_move[1][1]][last_move[1][0]]
+            if (piece_type == 6 or piece_type == 12) and self.color(last_move[1]) == inverse_color(color):
                 if abs(last_move[1][1] - last_move[0][1]) == 2 and abs(last_move[0][0] - f) == 1 and last_move[1][1] == r:
                     if color == 2:
-                        yield (last_move[0][0], r - 1, False, self.get_coord((f, r)))
+                        yield (last_move[0][0], r - 1, False, self.board[r][f])
                     else:
-                        yield (last_move[0][0], r + 1, False, self.get_coord((f, r)))
+                        yield (last_move[0][0], r + 1, False, self.board[r][f])
 
 
-    def rook(self, f, r):
+    def __rook(self, f, r):
         """Returns all valid moves for a rook at file f and rank r."""
-        return self.orthogonal([], f, r)
+        return self.__orthogonal([], f, r)
 
 
-    def bishop(self, f, r):
+    def __bishop(self, f, r):
         """Returns all valid moves for a bishop at file f and rank r."""
-        return self.diagonal([], f, r)
+        return self.__diagonal([], f, r)
 
 
-    def knight(self, f, r):
+    def __knight(self, f, r):
         """Returns all valid moves for a knight at file f and rank r."""
         return self.__warps([], (f, r), knight_warps, (self.color((f, r)),))
 
 
-    def queen(self, f, r):
+    def __queen(self, f, r):
         """Returns all valid moves for a queen at file f and rank r."""
-        return self.orthogonal(self.diagonal([], f, r), f, r)
+        return self.__orthogonal(self.__diagonal([], f, r), f, r)
 
 
-    def king(self, f, r, threats=None):
+    def __king(self, f, r, threats=None):
         """Returns all valid moves for a king at file f and rank r including castling (does not allow king to move into check or checkmate)."""
         color = self.color((f, r))
         castle_moves = []
         if self.kings_moved[color - 1] == -1:
-            #if self.rooks_moved[0 if color == WHITE else 2] == -1 and self.positions_clear([(1, r), (2, r), (3, r)], threats): #king does not move through (1, r)?
-            if self.rooks_moved[0 if color == WHITE else 2] == -1 and self.positions_clear([(2, r), (3, r)], threats):
+            if self.rooks_moved[0 if color == WHITE else 2] == -1 and self.__positions_clear([(2, r), (3, r)], threats):
                     castle_moves.append((2, r, False))
-            if self.rooks_moved[1 if color == WHITE else 3] == -1 and self.positions_clear([(5, r), (6, r)], threats):
+            if self.rooks_moved[1 if color == WHITE else 3] == -1 and self.__positions_clear([(5, r), (6, r)], threats):
                 castle_moves.append((6, r, False))
         for move in self.__warps(castle_moves, (f, r), king_warps, (color,)):
             yield move
 
 
-    def positions_clear(self, coords, threats=None):
+    def __positions_clear(self, coords, threats=None):
         """
         Determines if all coordinates in coords list are clear of both pieces and threats.
 
@@ -460,21 +484,22 @@ class Chess:
         """
         for coord in coords:
             status = 0 if threats is None else threats[coord[1]][coord[0]]
-            if self.get_coord(coord) != 0 or status != 0:
+            if self.board[coord[1]][coord[0]] != 0 or status != 0:
                 return False
         return True
 
 
     def __warps(self, moves, coord, relative_warps, excluded, threat=True):
+        #Warps can result in capturing the piece at the destination, so threat is True
         for move in moves:
             yield move
         for w in relative_warps:
             warp = (coord[0] + w[0], coord[1] + w[1], threat)
-            if in_range(warp) and self.color(warp) not in excluded:
+            if warp[0] < 8 and warp[0] >= 0 and warp[1] < 8 and warp[1] >= 0 and self.color(warp) not in excluded:
                 yield warp
 
 
-    def orthogonal(self, moves, f, r):
+    def __orthogonal(self, moves, f, r):
         """
         Compute move set for a piece that has orthogonal (horizontal and vertical) movement.
 
@@ -495,7 +520,7 @@ class Chess:
         return self.__vectors(moves, f, r, [(1, 0), (-1, 0), (0, 1), (0, -1)])
 
 
-    def diagonal(self, moves, f, r):
+    def __diagonal(self, moves, f, r):
         """
         Compute move set for a piece that has orthogonal (horizontal and vertical) movement.
 
@@ -514,7 +539,7 @@ class Chess:
             The input list of moves with diagonal moves appended to it.
         """
         return self.__vectors(moves, f, r, [(1, 1), (-1, -1), (1, -1), (-1, 1)])
-
+    
 
     def __vectors(self, moves, f, r, offsets_list):
         """
@@ -541,44 +566,20 @@ class Chess:
         """
         for move in moves:
             yield move
-        for offset in offsets_list:
-            for move in self.__vector(f, r, offset):
-                yield move
-
-
-    def __vector(self, f, r, offset):
-        """
-        Generates moves going in a single vector from a starting position until
-        either the edge of the board or another piece is encountered.
-
-        Parameters
-        ----------
-        f: int
-            The current file of the piece
-        r: int
-            The current rank of the piece
-        offset: tuple
-            A vector to travel in (e.g. if start is (0,0) and vector is (1,1)
-            move set will be (0,0), (1,1), (2,2), (3,3), etc.
-            
-        Returns
-        _______
-        moves: list of tuples
-            The input list of moves with vectored moves appended to it.
-        """
         color = self.color((f, r))
         inv_color = inverse_color(color)
-        coord = (f + offset[1], r + offset[0], True)
-        while in_range(coord) and self.color(coord) in (inv_color, EMPTY):
-            yield coord
-            if self.color(coord) == inv_color:
-                break
-            coord = (offset[1] + coord[0], offset[0] + coord[1], coord[2])
+        for offset in offsets_list:
+            coord = (f + offset[1], r + offset[0], True)
+            while coord[0] < 8 and coord[0] >= 0 and coord[1] < 8 and coord[1] >= 0 and self.color(coord) in (inv_color, EMPTY):
+                yield coord
+                if self.color(coord) == inv_color:
+                    break
+                coord = (offset[1] + coord[0], offset[0] + coord[1], coord[2])
 
 
     def __type_in_coords(self, coords, type):
         for coord in coords:
-            if self.get_coord(coord) == type:
+            if self.board[coord[1]][coord[0]] == type:
                 yield coord
 
 
@@ -658,5 +659,6 @@ class Chess:
         """Get the types of all captured pieces."""
         capture_pieces = []
         for move_num in range(len(self.move_list)):
-            capture_pieces.append(self.captured_pieces[move_num][0])
+            if move_num in self.captured_pieces:
+                capture_pieces.append(self.captured_pieces[move_num][0])
         return capture_pieces
\ No newline at end of file
